#include <algebra/EigenvalueSolver.hpp>
#include <cmath>
#include <utils/pugs_config.hpp>
#ifdef PUGS_HAS_SLEPC
#include <algebra/PETScUtils.hpp>
#include <slepc.h>
#include <slepceps.h>
#endif   // PUGS_HAS_SLEPC

#ifdef PUGS_HAS_EIGEN3
#include <algebra/Eigen3Utils.hpp>
#endif   // PUGS_HAS_EIGEN3

struct EigenvalueSolver::Internals
{
#ifdef PUGS_HAS_SLEPC
  static PetscReal
  computeSmallestRealEigenvalueOfSymmetricMatrix(EPS& eps)
  {
    EPSSetWhichEigenpairs(eps, EPS_SMALLEST_REAL);
    EPSSolve(eps);

    PetscReal smallest_eigenvalue;
    EPSGetEigenpair(eps, 0, &smallest_eigenvalue, nullptr, nullptr, nullptr);
    return smallest_eigenvalue;
  }

  static PetscReal
  computeLargestRealEigenvalueOfSymmetricMatrix(EPS& eps)
  {
    EPSSetWhichEigenpairs(eps, EPS_LARGEST_REAL);
    EPSSolve(eps);

    PetscReal largest_eigenvalue;
    EPSGetEigenpair(eps, 0, &largest_eigenvalue, nullptr, nullptr, nullptr);
    return largest_eigenvalue;
  }

  static void
  computeAllEigenvaluesOfSymmetricMatrixInInterval(EPS& eps, const PetscReal left_bound, const PetscReal right_bound)
  {
    Assert(left_bound < right_bound);
    EPSSetType(eps, EPSKRYLOVSCHUR);
    EPSSetWhichEigenpairs(eps, EPS_ALL);
    EPSSetInterval(eps, left_bound - 0.01 * std::abs(left_bound), right_bound + 0.01 * std::abs(right_bound));

    ST st;
    EPSGetST(eps, &st);
    STSetType(st, STSINVERT);

    KSP ksp;
#if (SLEPC_VERSION_MAJOR >= 3) && (SLEPC_VERSION_MINOR >= 15)
    EPSKrylovSchurGetKSP(eps, &ksp);
#else
    STGetKSP(st, &ksp);
#endif
    KSPSetType(ksp, KSPPREONLY);

    PC pc;
    KSPGetPC(ksp, &pc);
    PCSetType(pc, PCCHOLESKY);
    EPSSetFromOptions(eps);

    EPSSolve(eps);
  }

  static void
  computeAllEigenvaluesOfSymmetricMatrix(EPS& eps)
  {
    const PetscReal left_bound  = computeSmallestRealEigenvalueOfSymmetricMatrix(eps);
    const PetscReal right_bound = computeLargestRealEigenvalueOfSymmetricMatrix(eps);

    computeAllEigenvaluesOfSymmetricMatrixInInterval(eps, left_bound - 0.01 * std::abs(left_bound),
                                                     right_bound + 0.01 * std::abs(right_bound));
  }

  static void
  slepscComputeForSymmetricMatrix(const PETScAijMatrixEmbedder& A, SmallArray<double>& eigenvalues)
  {
    EPS eps;

    EPSCreate(PETSC_COMM_SELF, &eps);
    EPSSetOperators(eps, A, nullptr);
    EPSSetProblemType(eps, EPS_HEP);

    Internals::computeAllEigenvaluesOfSymmetricMatrix(eps);

    PetscInt nb_eigenvalues;
    EPSGetDimensions(eps, &nb_eigenvalues, nullptr, nullptr);

    eigenvalues = SmallArray<double>(nb_eigenvalues);
    for (PetscInt i = 0; i < nb_eigenvalues; ++i) {
      EPSGetEigenpair(eps, i, &(eigenvalues[i]), nullptr, nullptr, nullptr);
    }

    EPSDestroy(&eps);
  }

  static void
  slepscComputeForSymmetricMatrix(const PETScAijMatrixEmbedder& A,
                                  SmallArray<double>& eigenvalues,
                                  std::vector<SmallVector<double>>& eigenvector_list)
  {
    EPS eps;

    EPSCreate(PETSC_COMM_SELF, &eps);
    EPSSetOperators(eps, A, nullptr);
    EPSSetProblemType(eps, EPS_HEP);

    Internals::computeAllEigenvaluesOfSymmetricMatrix(eps);

    PetscInt nb_eigenvalues;
    EPSGetDimensions(eps, &nb_eigenvalues, nullptr, nullptr);

    eigenvalues = SmallArray<double>(nb_eigenvalues);
    eigenvector_list.reserve(nb_eigenvalues);
    for (PetscInt i = 0; i < nb_eigenvalues; ++i) {
      Vec Vr;
      SmallVector<double> eigenvector{A.numberOfRows()};
      VecCreateSeqWithArray(PETSC_COMM_SELF, 1, A.numberOfRows(), &(eigenvector[0]), &Vr);
      EPSGetEigenpair(eps, i, &(eigenvalues[i]), nullptr, Vr, nullptr);
      VecDestroy(&Vr);
      eigenvector_list.push_back(eigenvector);
    }

    EPSDestroy(&eps);
  }

  static void
  slepscComputeForSymmetricMatrix(const PETScAijMatrixEmbedder& A,
                                  SmallArray<double>& eigenvalues,
                                  SmallMatrix<double>& P)
  {
    EPS eps;

    EPSCreate(PETSC_COMM_SELF, &eps);
    EPSSetOperators(eps, A, nullptr);
    EPSSetProblemType(eps, EPS_HEP);

    Internals::computeAllEigenvaluesOfSymmetricMatrix(eps);

    PetscInt nb_eigenvalues;
    EPSGetDimensions(eps, &nb_eigenvalues, nullptr, nullptr);

    eigenvalues = SmallArray<double>(nb_eigenvalues);
    P           = SmallMatrix<double>(nb_eigenvalues, nb_eigenvalues);

    Array<double> eigenvector(nb_eigenvalues);
    for (PetscInt i = 0; i < nb_eigenvalues; ++i) {
      Vec Vr;
      VecCreateSeqWithArray(PETSC_COMM_SELF, 1, A.numberOfRows(), &(eigenvector[0]), &Vr);
      EPSGetEigenpair(eps, i, &(eigenvalues[i]), nullptr, Vr, nullptr);
      VecDestroy(&Vr);
      for (size_t j = 0; j < eigenvector.size(); ++j) {
        P(j, i) = eigenvector[j];
      }
    }

    EPSDestroy(&eps);
  }
#endif   // PUGS_HAS_SLEPC

#ifdef PUGS_HAS_EIGEN3

  template <typename Eigen3MatrixEmbedderType>
  static void
  eigen3ComputeForSymmetricMatrix(const Eigen3MatrixEmbedderType& A, SmallArray<double>& eigenvalues)
  {
    using Eigen3MatrixType = typename Eigen3MatrixEmbedderType::Eigen3MatrixType;
    Eigen::SelfAdjointEigenSolver<Eigen3MatrixType> eigen_solver;

    eigen_solver.compute(A.matrix(), Eigen::EigenvaluesOnly);

    eigenvalues = SmallArray<double>(A.numberOfRows());
    for (size_t i = 0; i < eigenvalues.size(); ++i) {
      eigenvalues[i] = eigen_solver.eigenvalues()[i];
    }
  }

  template <typename Eigen3MatrixEmbedderType>
  static void
  eigen3ComputeForSymmetricMatrix(const Eigen3MatrixEmbedderType& A,
                                  SmallArray<double>& eigenvalues,
                                  std::vector<SmallVector<double>>& eigenvectors)
  {
    using Eigen3MatrixType = typename Eigen3MatrixEmbedderType::Eigen3MatrixType;
    Eigen::SelfAdjointEigenSolver<Eigen3MatrixType> eigen_solver;

    eigen_solver.compute(A.matrix(), Eigen::ComputeEigenvectors);

    eigenvalues = SmallArray<double>(A.numberOfRows());
    for (size_t i = 0; i < eigenvalues.size(); ++i) {
      eigenvalues[i] = eigen_solver.eigenvalues()[i];
    }

    eigenvectors.resize(eigenvalues.size());
    for (size_t i = 0; i < eigenvalues.size(); ++i) {
      eigenvectors[i] = SmallVector<double>{eigenvalues.size()};
      for (size_t j = 0; j < eigenvalues.size(); ++j) {
        eigenvectors[i][j] = eigen_solver.eigenvectors().coeff(j, i);
      }
    }
  }

  template <typename Eigen3MatrixEmbedderType>
  static void
  eigen3ComputeForSymmetricMatrix(const Eigen3MatrixEmbedderType& A,
                                  SmallArray<double>& eigenvalues,
                                  SmallMatrix<double>& P)
  {
    using Eigen3MatrixType = typename Eigen3MatrixEmbedderType::Eigen3MatrixType;
    Eigen::SelfAdjointEigenSolver<Eigen3MatrixType> eigen_solver;

    eigen_solver.compute(A.matrix(), Eigen::ComputeEigenvectors);

    eigenvalues = SmallArray<double>(A.numberOfRows());
    for (size_t i = 0; i < eigenvalues.size(); ++i) {
      eigenvalues[i] = eigen_solver.eigenvalues()[i];
    }

    P = SmallMatrix<double>(eigenvalues.size(), eigenvalues.size());
    for (size_t i = 0; i < eigenvalues.size(); ++i) {
      for (size_t j = 0; j < eigenvalues.size(); ++j) {
        P(i, j) = eigen_solver.eigenvectors().coeff(i, j);
      }
    }
  }
#endif   // PUGS_HAS_EIGEN3
};

#ifdef PUGS_HAS_SLEPC
void
EigenvalueSolver::_slepscComputeForSymmetricMatrix(const CRSMatrix<double>& A, SmallArray<double>& eigenvalues)
{
  Internals::slepscComputeForSymmetricMatrix(A, eigenvalues);
}
#endif   // PUGS_HAS_SLEPC

template <typename T>
PUGS_INLINE TinyMatrix<3, 3, T>
EigenvalueSolver::swap(TinyMatrix<3, 3, T>& matrix, size_t i, size_t j) const
{
  TinyMatrix<3, 3, T> swapMat = matrix;
  for (size_t k = 0; k < 3; k++) {
    double valuei = matrix(i, k);
    swapMat(i, k) = matrix(j, k);
    swapMat(j, k) = valuei;
  }
  return swapMat;
}

template <typename T>
PUGS_INLINE constexpr TinyMatrix<3, 3, T>
EigenvalueSolver::rowReduce(const TinyMatrix<3, 3, T>& matrix, const double epsilon) const
{
  TinyMatrix<3, 3> redmat = matrix;
  for (size_t i = 0; i < 3; ++i) {
    size_t pivotRow = i;
    while (pivotRow < 3 && std::fabs(redmat(pivotRow, i)) < epsilon) {
      ++pivotRow;
    }
    if (pivotRow == 3) {
      continue;
    }
    // std::cout << "before: "
    //           << "i = " << i << " pivotRow = " << pivotRow << " " << redmat << "\n";
    redmat = swap(redmat, i, pivotRow);
    // std::cout << "after: " << redmat << "\n";

    double pivotValue = redmat(i, i);
    for (size_t j = i; j < 3; ++j) {
      redmat(i, j) /= pivotValue;
    }

    for (size_t k = i + 1; k < 3; ++k) {
      double factor = redmat(k, i);
      for (size_t j = i; j < 3; ++j) {
        redmat(k, j) -= factor * redmat(i, j);
      }
    }
  }
  return redmat;
}

TinyMatrix<3, 3>
EigenvalueSolver::findEigenvector(const TinyMatrix<3, 3>& A,
                                  const double eigenvalue,
                                  const size_t space_size,
                                  const double epsilon) const
{
  TinyMatrix<3, 3> eigenvectors = zero;
  if (space_size == 3) {
    // std::cout << "space_size = 3 "
    //           << "\n";
    return eigenvectors;
  }
  TinyMatrix<3, 3> B = A;

  for (size_t i = 0; i < 3; ++i) {
    B(i, i) -= eigenvalue;
  }

  // std::cout << " Avant " << B << "\n"
  //           << "\n";

  B = rowReduce(B, epsilon);

  // std::cout << " Après " << B << "\n"
  //           << "\n";
  for (size_t i = 0; i < 3; ++i) {
    bool isFreeVariableRow = true;
    for (size_t j = 0; j < 3; ++j) {
      if (std::fabs(B(i, j)) > epsilon) {
        isFreeVariableRow = false;
        // std::cout << " B(" << i << "," << j << ") = " << B(i, j) << "\n";
        break;
      }
    }

    if (isFreeVariableRow) {
      TinyVector<3> eigenvector = zero;
      eigenvector[i]            = 1.0;
      // std::cout << i << " is free "
      //           << "\n";
      // std::cout << std::flush;
      for (int k = i - 1; k >= 0; --k) {
        double sum = 0.0;
        for (size_t j = k + 1; j < 3; ++j) {
          sum += B(k, j) * eigenvector[j];
        }
        eigenvector[k] = -sum;
      }

      eigenvector = 1. / l2Norm(eigenvector) * eigenvector;

      for (size_t j = 0; j < 3; ++j) {
        eigenvectors(i, j) = eigenvector[j];
      }
    }
  }
  // std::cout << " Before orthorgonalization " << B << "\n";
  // std::cout << " Before orthorgonalization " << eigenvectors << "\n";

  if (space_size == 2) {
    TinyVector<3> first  = zero;
    TinyVector<3> second = zero;
    //      TinyVector<3> new_second =zero;
    // size_t rank1 = 0;
    size_t rank2 = 1;
    for (size_t j = 0; j < 3; ++j) {
      first[j]  = eigenvectors(0, j);
      second[j] = eigenvectors(1, j);
    }
    if (l2Norm(first) < 1e-3) {
      for (size_t j = 0; j < 3; ++j) {
        first[j] = eigenvectors(2, j);
        // rank1    = 2;
      }
    }
    if (l2Norm(second) < 1e-3) {
      for (size_t j = 0; j < 3; ++j) {
        second[j] = eigenvectors(2, j);
        rank2     = 2;
      }
    }
    double scalprod             = dot(first, second);
    double norm2                = dot(first, first);
    TinyVector<3> normal_vector = second - scalprod / norm2 * first;
    normal_vector               = 1. / l2Norm(normal_vector) * normal_vector;
    for (size_t j = 0; j < 3; ++j) {
      eigenvectors(rank2, j) = normal_vector[j];
    }
    // std::cout << " rank1 : " << rank1 << " rank2 " << rank2 << "\n";
  }
  // std::cout << "In findEigenvectors for eigenvalue " << eigenvalue << " eigenvectors " << eigenvectors << "\n";
  return eigenvectors;
}

bool
EigenvalueSolver::isDiagonal(const TinyMatrix<3, 3>& A, const double epsilon)
{
  bool isDiag = true;
  for (size_t i = 0; i < 2; i++) {
    for (size_t j = i + 1; j < 3; j++) {
      isDiag = !(fabs(A(i, j)) > epsilon);
      return isDiag;
    }
  }
  return isDiag;
}

TinyMatrix<3, 3>
EigenvalueSolver::findEigenvectors(const TinyMatrix<3, 3>& A, TinyVector<3> eigenvalues, const double epsilon) const
{
  TinyMatrix<3> eigenMatrix     = zero;
  size_t count                  = 0;
  TinyVector<3, int> space_size = zero;
  for (size_t i = 0; i < 3; i++) {
    space_size[i] = 1;
    for (size_t j = i + 1; j < 3; j++) {
      if (fabs(eigenvalues[i] - eigenvalues[j]) < epsilon) {
        double temp        = eigenvalues[i + 1];
        eigenvalues[i + 1] = eigenvalues[j];
        eigenvalues[j]     = temp;
        space_size[i] += 1;
        i += 1;
      }
    }
  }
  // std::cout << "space_size: " << space_size << "\n";
  // std::cout << "eigenvalues: " << eigenvalues << "\n";
  size_t it = 0;
  while (count < 3 && it < 3) {
    TinyMatrix<3> Try = findEigenvector(A, eigenvalues[count], space_size[count], epsilon);
    // std::cout << eigenvalues[count] << " Frobenius norm try: " << frobeniusNorm(Try) << "\n";
    Assert(frobeniusNorm(Try) > 1e-3, " Error : no eigenvector corresponds to eigenvalue ");
    if (frobeniusNorm(Try) < 1e-3) {
      std::cout << " Error : no eigenvector corresponds to eigenvalue \n";
      exit(0);
    }

    TinyVector<3> eigenvector = zero;
    int count2                = 0;
    for (size_t i = 0; i < 3; ++i) {
      for (size_t j = 0; j < 3; j++) {
        eigenvector[j] = Try(i, j);
      }
      // std::cout << " count, i: " << count << ", " << i << " l2Norm(eigenvector) " << l2Norm(eigenvector) << "\n";
      if (l2Norm(eigenvector) > 1e-3) {
        for (size_t j = 0; j < 3; j++) {
          eigenMatrix(count, j) = eigenvector[j];
        }
        count += 1;
        count2 += 1;
      }
    }
    Assert(count2 == space_size[count - count2], " eigenvector space size is not what was expected ");
    if (count2 != space_size[count - count2]) {
      std::cout << " eigenvector space size is not what was expected \n";
      exit(0);
    }
    it++;
  }
  // std ::cout << " eigenMatrix " << eigenMatrix << "\n";
  return eigenMatrix;
}

std::tuple<TinyVector<3>, TinyMatrix<3>>
EigenvalueSolver::findEigen(const TinyMatrix<3> A)
{
  constexpr TinyVector<3> eigenvalues0  = zero;
  constexpr TinyMatrix<3> eigenvectors0 = identity;
  const double epsilon                  = 1.e-6;   // * l2Norm(eigenvalues);
  if (frobeniusNorm(A) < 1.e-15) {
    // std::cout << " Frobenius norm 0 " << frobeniusNorm(A) << "\n";
    return std::make_tuple(eigenvalues0, eigenvectors0);
  }
  TinyMatrix<3> C = 1. / frobeniusNorm(A) * A;
  if (isDiagonal(C, epsilon)) {
    std::cout << "Matrix C " << C << " is diagonal "
              << "\n";
    const TinyVector<3> eigenvalues(A(0, 0), A(1, 1), A(2, 2));
    TinyMatrix<3> Diag = zero;
    for (size_t i = 0; i < 3; ++i) {
      Diag(i, i) = eigenvalues[i];
    }
    TinyMatrix<3> B = eigenvectors0 * Diag * transpose(eigenvectors0);
    // std::cout << "\n"
    //           << B << ", " << A << " Diff "
    //           << " A - B " << 1 / frobeniusNorm(A) * (A - B) << "\n";
    return std::make_tuple(eigenvalues, eigenvectors0);
  }
  const TinyVector<3> eigenvalues = _findEigenValues(C, epsilon);
  // std::cout << std::setprecision(15) << eigenvalues << "\n";
  TinyMatrix<3> Eigenmatrix = transpose(findEigenvectors(C, eigenvalues, epsilon));
  TinyMatrix<3> Diag        = zero;
  for (size_t i = 0; i < 3; ++i) {
    Diag(i, i) = frobeniusNorm(A) * eigenvalues[i];
  }
  TinyMatrix<3> B = Eigenmatrix * Diag * transpose(Eigenmatrix);
  // std::cout << "\n"
  //           << B << ", " << A << " Diff "
  //           << " A - B " << 1 / frobeniusNorm(A) * (A - B) << "\n";
  return std::make_tuple(frobeniusNorm(A) * eigenvalues, Eigenmatrix);
}
TinyVector<3>
EigenvalueSolver::_findEigenValues(const TinyMatrix<3>& A, const double epsilon) const
{
  // std::cout << " Matrix " << A << "\n";
  TinyVector<3> eigenvalues(0., 0., 0.);
  double b = -trace(A);
  double c = 0.5 * (trace(A) * trace(A) - trace(A * A));
  double d = -det(A);
  Polynomial<3> P(d, c, b, 1.);
  // std::cout << P << "\n";
  Polynomial<2> Q(c, b, 1.);
  if (fabs(d) > 1e-11) {
    eigenvalues[0] = _findFirstRoot(P);
    Polynomial<1> Q1(-eigenvalues[0], 1);
    Polynomial<3> S;
    Polynomial<3> R;
    divide(P, Q1, S, R);
    // std::cout << "Q1 : " << Q1 << "\n";
    // std::cout << "R : " << R << "\n";
    // std::cout << "S : " << S << "\n";
    Q.coefficients()[0] = S.coefficients()[0];
    Q.coefficients()[1] = S.coefficients()[1];
    Q.coefficients()[2] = S.coefficients()[2];
  }
  TinyVector<2> last_eigenvalues = _findLastRoots(Q, epsilon);
  eigenvalues[1]                 = last_eigenvalues[0];
  eigenvalues[2]                 = last_eigenvalues[1];
  TinyVector<3, int> space_size  = zero;
  for (size_t i = 0; i < 3; i++) {
    space_size[i] = 1;
    for (size_t j = i + 1; j < 3; j++) {
      if (eigenvalues[i] == eigenvalues[j]) {
        double temp        = eigenvalues[i + 1];
        eigenvalues[i + 1] = eigenvalues[j];
        eigenvalues[j]     = temp;
        space_size[i] += 1;
        i += 1;
      }
    }
  }
  return eigenvalues;
}

double
EigenvalueSolver::_findFirstRoot(Polynomial<3> P) const
{
  double guess     = -P.coefficients()[2] / 3.;
  double old_guess = -P.coefficients()[2] / 3.;
  // std::cout << " coefs P " << P.coefficients() << "\n";
  //    double delta = pow(2.*P.coefficients()[2],2) - 4*3.*P.coefficients()[3]*P.coefficients()[1];
  Polynomial<2> Q = derivative(P);
  Polynomial<1> R = derivative(Q);
  // size_t iteration = 0;
  double residu    = 0.;
  size_t iteration = 0;
  do {
    //      guess -= P(guess) / Q(guess);
    old_guess = guess;
    guess -= 2 * P(guess) * Q(guess) / (2 * Q(guess) * Q(guess) - P(guess) * R(guess));
    // std::cout << "guess = " << guess << " old_guess " << old_guess << "\n";
    //  std::cout << "g(guess) = " << Q(guess) << "\n";
    residu = P(guess);
    // std::cout << "residu = " << residu << " delta " << fabs(old_guess - guess) << "\n";
    iteration += 1;
  } while (((fabs(residu) > 1.e-16) or (fabs(old_guess - guess) > 1.e-10)) and (iteration < 10000));
  if (iteration == 100) {
    std::cout << " nb Newton iterations reached, residu " << residu << "\n";
  }
  return guess;
}

TinyVector<2>
EigenvalueSolver::_findLastRoots(Polynomial<2> P, const double epsilon) const
{
  TinyVector<2> roots(0., 0.);
  double delta = P.coefficients()[1] * P.coefficients()[1] - 4 * P.coefficients()[2] * P.coefficients()[0];
  // if (fabs(delta) > epsilon) {
  //   std::cout << "Find roots is only for symmetric matrix \n";
  //   exit(0);
  // }
  Assert(delta > -epsilon, "Find roots is only for symmetric matrix");
  if (fabs(delta) < epsilon) {
    roots[0] = -P.coefficients()[1] / (2. * P.coefficients()[2]);
    roots[1] = roots[0];
  }
  if (delta >= 0.) {
    roots[0] = (-P.coefficients()[1] - std::sqrt(delta)) / (2. * P.coefficients()[2]);
    roots[1] = (-P.coefficients()[1] + std::sqrt(delta)) / (2. * P.coefficients()[2]);
  }
  return roots;
}

TinyMatrix<3>
EigenvalueSolver::computeExpForTinyMatrix(const TinyMatrix<3>& A)
{
  TinyMatrix<3> expA;
  auto [eigenvalues, eigenvectors] = findEigen(A);
  TinyMatrix<3> D                  = zero;
  for (size_t i = 0; i < 3; ++i) {
    if (std::abs(eigenvalues[i]) > 200) {
      std::cout << "Warning: large eigenvalue for matrix " << A << " eigenvalues " << eigenvalues << "\n";
      //      eigenvalues[i] = eigenvalues[i] / eigenvalues[i] * 200.;
      //      exit(0);
    }
  }
  for (size_t i = 0; i < 3; ++i) {
    D(i, i) = std::exp(eigenvalues[i]);
  }
  expA = eigenvectors * D * transpose(eigenvectors);
  return expA;
}

#ifdef PUGS_HAS_SLEPC

void
EigenvalueSolver::_slepscComputeForSymmetricMatrix(const DenseMatrixWrapper<double>& A, SmallArray<double>& eigenvalues)
{
  Internals::slepscComputeForSymmetricMatrix(A, eigenvalues);
}

void
EigenvalueSolver::_slepscComputeForSymmetricMatrix(const CRSMatrix<double>& A,
                                                   SmallArray<double>& eigenvalues,
                                                   std::vector<SmallVector<double>>& eigenvectors)
{
  Internals::slepscComputeForSymmetricMatrix(A, eigenvalues, eigenvectors);
}

void
EigenvalueSolver::_slepscComputeForSymmetricMatrix(const DenseMatrixWrapper<double>& A,
                                                   SmallArray<double>& eigenvalues,
                                                   std::vector<SmallVector<double>>& eigenvectors)
{
  Internals::slepscComputeForSymmetricMatrix(A, eigenvalues, eigenvectors);
}

void
EigenvalueSolver::_slepscComputeForSymmetricMatrix(const CRSMatrix<double>& A,
                                                   SmallArray<double>& eigenvalues,
                                                   SmallMatrix<double>& P)
{
  Internals::slepscComputeForSymmetricMatrix(A, eigenvalues, P);
}

void
EigenvalueSolver::_slepscComputeForSymmetricMatrix(const DenseMatrixWrapper<double>& A,
                                                   SmallArray<double>& eigenvalues,
                                                   SmallMatrix<double>& P)
{
  Internals::slepscComputeForSymmetricMatrix(A, eigenvalues, P);
}

#else   // PUGS_HAS_SLEPC

void
EigenvalueSolver::_slepscComputeForSymmetricMatrix(const CRSMatrix<double>&, SmallArray<double>&)
{}

void
EigenvalueSolver::_slepscComputeForSymmetricMatrix(const DenseMatrixWrapper<double>&, SmallArray<double>&)
{}

void
EigenvalueSolver::_slepscComputeForSymmetricMatrix(const CRSMatrix<double>&,
                                                   SmallArray<double>&,
                                                   std::vector<SmallVector<double>>&)
{}

void
EigenvalueSolver::_slepscComputeForSymmetricMatrix(const DenseMatrixWrapper<double>&,
                                                   SmallArray<double>&,
                                                   std::vector<SmallVector<double>>&)
{}

void
EigenvalueSolver::_slepscComputeForSymmetricMatrix(const CRSMatrix<double>&, SmallArray<double>&, SmallMatrix<double>&)
{}

void
EigenvalueSolver::_slepscComputeForSymmetricMatrix(const DenseMatrixWrapper<double>&,
                                                   SmallArray<double>&,
                                                   SmallMatrix<double>&)
{}

#endif   // PUGS_HAS_SLEPC

#ifdef PUGS_HAS_EIGEN3
void
EigenvalueSolver::_eigen3ComputeForSymmetricMatrix(const CRSMatrix<double>& A, SmallArray<double>& eigenvalues)
{
  Eigen3SparseMatrixEmbedder embedded_A{A};
  Internals::eigen3ComputeForSymmetricMatrix(embedded_A, eigenvalues);
}

void
EigenvalueSolver::_eigen3ComputeForSymmetricMatrix(const DenseMatrixWrapper<double>& A, SmallArray<double>& eigenvalues)
{
  Internals::eigen3ComputeForSymmetricMatrix(Eigen3DenseMatrixEmbedder{A}, eigenvalues);
}

void
EigenvalueSolver::_eigen3ComputeForSymmetricMatrix(const CRSMatrix<double>& A,
                                                   SmallArray<double>& eigenvalues,
                                                   std::vector<SmallVector<double>>& eigenvectors)
{
  Internals::eigen3ComputeForSymmetricMatrix(Eigen3SparseMatrixEmbedder{A}, eigenvalues, eigenvectors);
}

void
EigenvalueSolver::_eigen3ComputeForSymmetricMatrix(const DenseMatrixWrapper<double>& A,
                                                   SmallArray<double>& eigenvalues,
                                                   std::vector<SmallVector<double>>& eigenvectors)
{
  Internals::eigen3ComputeForSymmetricMatrix(Eigen3DenseMatrixEmbedder{A}, eigenvalues, eigenvectors);
}

void
EigenvalueSolver::_eigen3ComputeForSymmetricMatrix(const CRSMatrix<double>& A,
                                                   SmallArray<double>& eigenvalues,
                                                   SmallMatrix<double>& P)
{
  Internals::eigen3ComputeForSymmetricMatrix(Eigen3SparseMatrixEmbedder{A}, eigenvalues, P);
}

void
EigenvalueSolver::_eigen3ComputeForSymmetricMatrix(const DenseMatrixWrapper<double>& A,
                                                   SmallArray<double>& eigenvalues,
                                                   SmallMatrix<double>& P)
{
  Internals::eigen3ComputeForSymmetricMatrix(Eigen3DenseMatrixEmbedder{A}, eigenvalues, P);
}

#else   // PUGS_HAS_EIGEN3

void
EigenvalueSolver::_eigen3ComputeForSymmetricMatrix(const CRSMatrix<double>&, SmallArray<double>&)
{}

void
EigenvalueSolver::_eigen3ComputeForSymmetricMatrix(const DenseMatrixWrapper<double>&, SmallArray<double>&)
{}

void
EigenvalueSolver::_eigen3ComputeForSymmetricMatrix(const CRSMatrix<double>&,
                                                   SmallArray<double>&,
                                                   std::vector<SmallVector<double>>&)
{}

void
EigenvalueSolver::_eigen3ComputeForSymmetricMatrix(const DenseMatrixWrapper<double>&,
                                                   SmallArray<double>&,
                                                   std::vector<SmallVector<double>>&)
{}

void
EigenvalueSolver::_eigen3ComputeForSymmetricMatrix(const CRSMatrix<double>&, SmallArray<double>&, SmallMatrix<double>&)
{}

void
EigenvalueSolver::_eigen3ComputeForSymmetricMatrix(const DenseMatrixWrapper<double>&,
                                                   SmallArray<double>&,
                                                   SmallMatrix<double>&)
{}

#endif   // PUGS_HAS_EIGEN3

bool
EigenvalueSolver::hasLibrary(const ESLibrary library)
{
  switch (library) {
  case ESLibrary::eigen3: {
#ifdef PUGS_HAS_EIGEN3
    return true;
#else
    return false;
#endif
  }
  case ESLibrary::slepsc: {
#ifdef PUGS_HAS_SLEPC
    return true;
#else
    return false;
#endif
  }
    // LCOV_EXCL_START
  default: {
    throw UnexpectedError("Eigenvalue  library (" + ::name(library) + ") was not set!");
  }
    // LCOV_EXCL_STOP
  }
}

void
EigenvalueSolver::checkHasLibrary(const ESLibrary library)
{
  if (not hasLibrary(library)) {
    // LCOV_EXCL_START
    throw NormalError(::name(library) + " is not linked to pugs. Cannot use it!");
    // LCOV_EXCL_STOP
  }
}

EigenvalueSolver::EigenvalueSolver(const EigenvalueSolverOptions& options) : m_options{options}
{
  checkHasLibrary(options.library());
}
