#include <language/algorithms/ODEDiscontinuousGalerkin1D.hpp>
#include <language/utils/InterpolateArray.hpp>
#include <scheme/ODEGD1D.hpp>

#include <fstream>

template <size_t Dimension, size_t Degree>
ODEDiscontinuousGalerkin1D<Dimension, Degree>::ODEDiscontinuousGalerkin1D(const BasisType& basis_type,
                                                                          std::shared_ptr<const IMesh> i_mesh,
                                                                          const FunctionSymbolId& uex_id)
{
  using ConnectivityType = Connectivity<Dimension>;
  using MeshType         = Mesh<ConnectivityType>;

  std::shared_ptr<const MeshType> mesh = std::dynamic_pointer_cast<const MeshType>(i_mesh);
  static_assert(Dimension == 1, "Dimension have to be 1 for DiscontinuousGalerkin1D");
  using R1 = TinyVector<Dimension>;
  ODEDG1D<Degree> odedg1d(basis_type, mesh);
  const NodeValue<const R1> xr    = mesh->xr();
  const auto& cell_to_node_matrix = mesh->connectivity().cellToNodeMatrix();
  Array<double> my_array(100 * mesh->numberOfCells() + 1);

  for (CellId j = 0; j < mesh->numberOfCells(); ++j) {
    const auto& cell_nodes          = cell_to_node_matrix[j];
    const NodeId r1                 = cell_nodes[0];
    const NodeId r2                 = cell_nodes[1];
    const TinyVector<Dimension> xr1 = xr[r1];
    const TinyVector<Dimension> xr2 = xr[r2];
    const double deltax             = (xr2[0] - xr1[0]) / 100;
    bool start                      = (j == 0);
    for (size_t i = (!start); i <= 100; ++i) {
      const double x        = xr1[0] + deltax * i;
      my_array[100 * j + i] = x;
    }
  }

  Array<double> ej = InterpolateArray<double(double)>::interpolate(uex_id, my_array);
  // Array<TinyVector<Dimension>> ej = InterpolateArray<TinyVector<Dimension>(double)>::interpolate(uex_id, my_array);

  // InterpolateItemValue<double(TinyVector<Dimension>)>::template interpolate<ItemType::cell>(uex_id, mesh_data.xj());

  CellValue<Polynomial<Degree>> U(mesh->connectivity());
  odedg1d.globalSolve(U);
  PolynomialBasis<Degree> B;
  TinyVector<Degree + 1> zeros;
  for (size_t i = 0; i <= Degree; i++) {
    zeros[i] = i;
  }
  B.build(basis_type, 0.5, zeros);
  std::string filename = "result-basis-";
  filename += B.displayType();
  filename += "-degree-";
  filename += std::to_string(Degree);
  filename += "-dof-";
  filename += std::to_string(mesh->numberOfCells());
  std::string filename2 = filename + "-polynomials";
  std::ofstream sout(filename.c_str());
  std::ofstream sout2(filename2.c_str());
  for (CellId j = 0; j < mesh->numberOfCells(); ++j) {
    const auto& cell_nodes          = cell_to_node_matrix[j];
    const NodeId r1                 = cell_nodes[0];
    const NodeId r2                 = cell_nodes[1];
    const TinyVector<Dimension> xr1 = xr[r1];
    const TinyVector<Dimension> xr2 = xr[r2];
    const double deltax             = (xr2[0] - xr1[0]) / 100;
    bool start                      = (j == 0);
    double norm1error               = 0.;
    for (size_t i = (!start); i <= 100; ++i) {
      const double x = xr1[0] + deltax * i;
      sout2 << x << " " << U[j](x) << " " << ej[100 * j + i] << " " << std::abs(U[j](x) - ej[100 * j + i]) << '\n';
      norm1error += std::abs(U[j](x) - ej[100 * j + i]);
    }
    double x13 = (2 * xr1[0] + xr2[0]) / 3;
    double x23 = (xr1[0] + 2 * xr2[0]) / 3;

    double uex = std::exp(xr1[0]) + 3 * (std::exp(x13) + std::exp(x23)) + std::exp(xr2[0]);
    uex /= 8;
    double ucalc = integrate(U[j], xr1[0], xr2[0]) / (xr2[0] - xr1[0]);
    sout << (0.5 * (xr1[0] + xr2[0])) << " " << uex << " " << ucalc << " " << std::abs(ucalc - uex) << " " << norm1error
         << '\n';
  }
}

template ODEDiscontinuousGalerkin1D<1, 0>::ODEDiscontinuousGalerkin1D(const BasisType& basis_type,
                                                                      std::shared_ptr<const IMesh>,
                                                                      const FunctionSymbolId&);
template ODEDiscontinuousGalerkin1D<1, 1>::ODEDiscontinuousGalerkin1D(const BasisType& basis_type,
                                                                      std::shared_ptr<const IMesh>,
                                                                      const FunctionSymbolId&);
template ODEDiscontinuousGalerkin1D<1, 2>::ODEDiscontinuousGalerkin1D(const BasisType& basis_type,
                                                                      std::shared_ptr<const IMesh>,
                                                                      const FunctionSymbolId&);
template ODEDiscontinuousGalerkin1D<1, 3>::ODEDiscontinuousGalerkin1D(const BasisType& basis_type,
                                                                      std::shared_ptr<const IMesh>,
                                                                      const FunctionSymbolId&);
template ODEDiscontinuousGalerkin1D<1, 4>::ODEDiscontinuousGalerkin1D(const BasisType& basis_type,
                                                                      std::shared_ptr<const IMesh>,
                                                                      const FunctionSymbolId&);
