#ifndef STENCIL_ARRAY_HPP
#define STENCIL_ARRAY_HPP

#include <mesh/ConnectivityMatrix.hpp>
#include <mesh/IBoundaryDescriptor.hpp>
#include <mesh/ItemId.hpp>
#include <mesh/ItemToItemMatrix.hpp>

template <ItemType source_item_type, ItemType target_item_type>
class StencilArray
{
 public:
  using ItemToItemMatrixT = ItemToItemMatrix<source_item_type, target_item_type>;

  using SourceItemId = ItemIdT<source_item_type>;
  using TargetItemId = ItemIdT<target_item_type>;

  class BoundaryDescriptorStencilArray
  {
   private:
    std::shared_ptr<const IBoundaryDescriptor> m_iboundary_descriptor;
    const ConnectivityMatrix m_connectivity_matrix;
    const ItemToItemMatrixT m_stencil_array;

   public:
    const IBoundaryDescriptor&
    boundaryDescriptor() const
    {
      return *m_iboundary_descriptor;
    }

    const auto&
    stencilArray() const
    {
      return m_stencil_array;
    }

    BoundaryDescriptorStencilArray(const std::shared_ptr<const IBoundaryDescriptor>& iboundary_descriptor,
                                   const ConnectivityMatrix& connectivity_matrix)
      : m_iboundary_descriptor{iboundary_descriptor},
        m_connectivity_matrix{connectivity_matrix},
        m_stencil_array{m_connectivity_matrix}
    {}

    BoundaryDescriptorStencilArray(const BoundaryDescriptorStencilArray& bdsa)
      : m_iboundary_descriptor{bdsa.m_iboundary_descriptor},
        m_connectivity_matrix{bdsa.m_connectivity_matrix},
        m_stencil_array{m_connectivity_matrix}
    {}

    BoundaryDescriptorStencilArray(BoundaryDescriptorStencilArray&& bdsa)
      : m_iboundary_descriptor{std::move(bdsa.m_iboundary_descriptor)},
        m_connectivity_matrix{std::move(bdsa.m_connectivity_matrix)},
        m_stencil_array{m_connectivity_matrix}
    {}

    ~BoundaryDescriptorStencilArray() = default;
  };

  using BoundaryDescriptorStencilArrayList = std::vector<BoundaryDescriptorStencilArray>;

 private:
  const ConnectivityMatrix m_connectivity_matrix;
  const ItemToItemMatrixT m_stencil_array;
  BoundaryDescriptorStencilArrayList m_symmetry_boundary_stencil_array_list;

 public:
  PUGS_INLINE
  const auto&
  symmetryBoundaryStencilArrayList() const
  {
    return m_symmetry_boundary_stencil_array_list;
  }

  PUGS_INLINE
  auto
  operator[](SourceItemId source_item_id) const
  {
    return m_stencil_array[source_item_id];
  }

  StencilArray(const ConnectivityMatrix& connectivity_matrix,
               const BoundaryDescriptorStencilArrayList& symmetry_boundary_descriptor_stencil_array_list)
    : m_connectivity_matrix{connectivity_matrix},
      m_stencil_array{m_connectivity_matrix},
      m_symmetry_boundary_stencil_array_list{symmetry_boundary_descriptor_stencil_array_list}
  {}

  StencilArray(const StencilArray& stencil_array)
    : m_connectivity_matrix{stencil_array.m_connectivity_matrix},
      m_stencil_array{m_connectivity_matrix},
      m_symmetry_boundary_stencil_array_list{stencil_array.m_symmetry_boundary_stencil_array_list}
  {}

  StencilArray(StencilArray&& stencil_array)
    : m_connectivity_matrix{std::move(stencil_array.m_connectivity_matrix)},
      m_stencil_array{m_connectivity_matrix},
      m_symmetry_boundary_stencil_array_list{std::move(stencil_array.m_symmetry_boundary_stencil_array_list)}
  {}

  ~StencilArray() = default;
};

using CellToCellStencilArray = StencilArray<ItemType::cell, ItemType::cell>;
using CellToFaceStencilArray = StencilArray<ItemType::cell, ItemType::face>;
using NodeToCellStencilArray = StencilArray<ItemType::node, ItemType::cell>;

#endif   // STENCIL_ARRAY_HPP
