#ifndef HYPERPLASTIC_SOLVER_O2_HPP
#define HYPERPLASTIC_SOLVER_O2_HPP

#include <mesh/MeshTraits.hpp>

#include <memory>
#include <tuple>
#include <vector>

class IBoundaryConditionDescriptor;
class MeshVariant;
class ItemValueVariant;
class SubItemValuePerItemVariant;
class DiscreteFunctionVariant;

double hyperplastic_dt(const std::shared_ptr<const DiscreteFunctionVariant>& c);

class HyperplasticSolverO2Handler
{
 public:
  enum class SolverType
  {
    Glace,
    Eucclhyd
  };

  enum class RelaxationType
  {
    Implicit,
    Exponential,
    CauchyGreen
  };

 private:
  struct IHyperplasticSolverO2
  {
    virtual std::tuple<const std::shared_ptr<const ItemValueVariant>,
                       const std::shared_ptr<const SubItemValuePerItemVariant>>
    compute_fluxes(
      const SolverType& solver_type,
      const std::shared_ptr<const DiscreteFunctionVariant>& rho,
      const std::shared_ptr<const DiscreteFunctionVariant>& aL,
      const std::shared_ptr<const DiscreteFunctionVariant>& aT,
      const std::shared_ptr<const DiscreteFunctionVariant>& u,
      const std::shared_ptr<const DiscreteFunctionVariant>& sigma,
      const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>& bc_descriptor_list) const = 0;

    virtual std::tuple<std::shared_ptr<const MeshVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>>
    apply_fluxes(const double& dt,
                 const std::shared_ptr<const DiscreteFunctionVariant>& rho,
                 const std::shared_ptr<const DiscreteFunctionVariant>& u,
                 const std::shared_ptr<const DiscreteFunctionVariant>& E,
                 const std::shared_ptr<const DiscreteFunctionVariant>& Fe,
                 const std::shared_ptr<const DiscreteFunctionVariant>& zeta,
                 const std::shared_ptr<const DiscreteFunctionVariant>& yield,
                 const std::shared_ptr<const DiscreteFunctionVariant>& sigma,
                 const std::shared_ptr<const ItemValueVariant>& ur,
                 const std::shared_ptr<const SubItemValuePerItemVariant>& Fjr,
                 const RelaxationType& relaxation_type,
                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>& bc_descriptor_list) const = 0;

    virtual std::tuple<std::shared_ptr<const MeshVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>>
    apply(const SolverType& solver_type,
          const RelaxationType& relaxation_type,
          const double& dt,
          const std::shared_ptr<const DiscreteFunctionVariant>& rho,
          const std::shared_ptr<const DiscreteFunctionVariant>& u,
          const std::shared_ptr<const DiscreteFunctionVariant>& E,
          const std::shared_ptr<const DiscreteFunctionVariant>& Fe,
          const std::shared_ptr<const DiscreteFunctionVariant>& zeta,
          const std::shared_ptr<const DiscreteFunctionVariant>& yield,
          const std::shared_ptr<const DiscreteFunctionVariant>& aL,
          const std::shared_ptr<const DiscreteFunctionVariant>& aT,
          const std::shared_ptr<const DiscreteFunctionVariant>& sigma,
          const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>& bc_descriptor_list) const = 0;

    virtual std::tuple<std::shared_ptr<const MeshVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>>
    apply_elastic(const SolverType& solver_type,
                  const double& dt,
                  const std::shared_ptr<const DiscreteFunctionVariant>& rho,
                  const std::shared_ptr<const DiscreteFunctionVariant>& u,
                  const std::shared_ptr<const DiscreteFunctionVariant>& E,
                  const std::shared_ptr<const DiscreteFunctionVariant>& Fe,
                  const std::shared_ptr<const DiscreteFunctionVariant>& aL,
                  const std::shared_ptr<const DiscreteFunctionVariant>& aT,
                  const std::shared_ptr<const DiscreteFunctionVariant>& sigma,
                  const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>& bc_descriptor_list) const = 0;

    virtual std::shared_ptr<const DiscreteFunctionVariant> apply_plastic_relaxation(
      const RelaxationType& relaxation_type,
      const double& dt,
      const std::shared_ptr<const DiscreteFunctionVariant>& Fe,
      const std::shared_ptr<const DiscreteFunctionVariant>& zeta,
      const std::shared_ptr<const DiscreteFunctionVariant>& yield,
      const std::shared_ptr<const DiscreteFunctionVariant>& sigman,
      const std::shared_ptr<const DiscreteFunctionVariant>& sigmanp1) const = 0;

    IHyperplasticSolverO2()                                   = default;
    IHyperplasticSolverO2(IHyperplasticSolverO2&&)            = default;
    IHyperplasticSolverO2& operator=(IHyperplasticSolverO2&&) = default;

    virtual ~IHyperplasticSolverO2() = default;
  };

  template <MeshConcept MeshType>
  class HyperplasticSolverO2;

  std::unique_ptr<IHyperplasticSolverO2> m_hyperplastic_solver;

 public:
  const IHyperplasticSolverO2&
  solver() const
  {
    return *m_hyperplastic_solver;
  }

  HyperplasticSolverO2Handler(const std::shared_ptr<const MeshVariant>& mesh);
};

#endif   // HYPERPLASTIC_SOLVER_O2_HPP
