#include <catch2/catch_approx.hpp>
#include <catch2/catch_test_macros.hpp>
#include <catch2/matchers/catch_matchers_all.hpp>

#include <Kokkos_Core.hpp>

#include <utils/PugsAssert.hpp>
#include <utils/Types.hpp>

#include <algebra/SmallMatrix.hpp>
#include <algebra/SmallVector.hpp>
#include <analysis/GaussLegendreQuadratureDescriptor.hpp>
#include <analysis/GaussQuadratureDescriptor.hpp>
#include <analysis/QuadratureFormula.hpp>
#include <analysis/QuadratureManager.hpp>
#include <geometry/CubeTransformation.hpp>
#include <geometry/LineTransformation.hpp>
#include <geometry/PrismTransformation.hpp>
#include <geometry/PyramidTransformation.hpp>
#include <geometry/SquareTransformation.hpp>
#include <geometry/TetrahedronTransformation.hpp>
#include <geometry/TriangleTransformation.hpp>
#include <mesh/Mesh.hpp>
#include <mesh/MeshDataManager.hpp>
#include <scheme/DiscreteFunctionDPkVariant.hpp>
#include <scheme/DiscreteFunctionP0.hpp>
#include <scheme/DiscreteFunctionVariant.hpp>
#include <scheme/PolynomialReconstruction.hpp>

#include <MeshDataBaseForTests.hpp>

// clazy:excludeall=non-pod-global-static

TEST_CASE("QuadraticPolynomialReconstruction", "[scheme]")
{
  SECTION("degree 2")
  {
    SECTION("1D")
    {
      PolynomialReconstructionDescriptor descriptor{IntegrationMethodType::element, 2};
      using R1 = TinyVector<1>;

      QuadratureFormula<1> qf = QuadratureManager::instance().getLineFormula(GaussLegendreQuadratureDescriptor{2});

      SECTION("R data")
      {
        for (auto named_mesh : MeshDataBaseForTests::get().all1DMeshes()) {
          SECTION(named_mesh.name())
          {
            auto p_mesh = named_mesh.mesh()->get<Mesh<1>>();
            auto& mesh  = *p_mesh;

            auto f_exact = [](const R1& x) { return 2.3 + 1.7 * x[0] - 3.2 * x[0] * x[0]; };

            auto xr = mesh.xr();
            auto Vj = MeshDataManager::instance().getMeshData(mesh).Vj();

            auto cell_to_node_matrix = mesh.connectivity().cellToNodeMatrix();

            DiscreteFunctionP0<double> fh{p_mesh};

            parallel_for(
              mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
                auto cell_node_list = cell_to_node_matrix[cell_id];
                double value        = 0;
                LineTransformation<1> T{xr[cell_node_list[0]], xr[cell_node_list[1]]};

                for (size_t i_q = 0; i_q < qf.numberOfPoints(); ++i_q) {
                  value += qf.weight(i_q) * f_exact(T(qf.point(i_q))) * T.jacobianDeterminant();
                }

                fh[cell_id] = value / Vj[cell_id];
              });

            auto reconstructions = PolynomialReconstruction{descriptor}.build(fh);

            auto dpk_fh = reconstructions[0]->get<DiscreteFunctionDPk<1, const double>>();

            {
              double L1_error = 0;
              for (CellId cell_id = 0; cell_id < mesh.numberOfCells(); ++cell_id) {
                const auto cell_node_list = cell_to_node_matrix[cell_id];
                const auto dpkj           = dpk_fh[cell_id];

                LineTransformation<1> T{xr[cell_node_list[0]], xr[cell_node_list[1]]};

                double error = 0;

                for (size_t i_q = 0; i_q < qf.numberOfPoints(); ++i_q) {
                  error += std::abs(qf.weight(i_q) * (f_exact(T(qf.point(i_q))) - dpkj(T(qf.point(i_q)))) *
                                    T.jacobianDeterminant());
                }

                L1_error += error;
              }
              REQUIRE(parallel::allReduceMax(L1_error) == Catch::Approx(0).margin(1E-13));
            }
          }
        }
      }

      SECTION("R^3 data")
      {
        using R3 = TinyVector<3>;

        for (auto named_mesh : MeshDataBaseForTests::get().all1DMeshes()) {
          SECTION(named_mesh.name())
          {
            auto p_mesh = named_mesh.mesh()->get<Mesh<1>>();
            auto& mesh  = *p_mesh;

            auto u_exact = [](const R1& X) -> R3 {
              const double x = X[0];

              return R3{2.3 + 1.7 * x - 3.2 * x * x,   //
                        7 + 5 * x + 3 * x * x,         //
                        -4 + 3.5 * x + 2.7 * x * x};
            };

            auto xr = mesh.xr();
            auto Vj = MeshDataManager::instance().getMeshData(mesh).Vj();

            auto cell_to_node_matrix = mesh.connectivity().cellToNodeMatrix();

            DiscreteFunctionP0<R3> uh{p_mesh};

            parallel_for(
              mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
                auto cell_node_list = cell_to_node_matrix[cell_id];
                R3 value            = zero;
                LineTransformation<1> T{xr[cell_node_list[0]], xr[cell_node_list[1]]};

                for (size_t i_q = 0; i_q < qf.numberOfPoints(); ++i_q) {
                  value += qf.weight(i_q) * T.jacobianDeterminant() * u_exact(T(qf.point(i_q)));
                }

                uh[cell_id] = 1. / Vj[cell_id] * value;
              });

            auto reconstructions = PolynomialReconstruction{descriptor}.build(uh);

            auto dpk_fh = reconstructions[0]->get<DiscreteFunctionDPk<1, const R3>>();

            {
              double L1_error = 0;
              for (CellId cell_id = 0; cell_id < mesh.numberOfCells(); ++cell_id) {
                const auto cell_node_list = cell_to_node_matrix[cell_id];
                const auto dpkj           = dpk_fh[cell_id];

                LineTransformation<1> T{xr[cell_node_list[0]], xr[cell_node_list[1]]};

                double error = 0;

                for (size_t i_q = 0; i_q < qf.numberOfPoints(); ++i_q) {
                  R3 diff =
                    qf.weight(i_q) * T.jacobianDeterminant() * (u_exact(T(qf.point(i_q))) - dpkj(T(qf.point(i_q))));
                  error += std::abs(diff[0]) + std::abs(diff[1]) + std::abs(diff[2]);
                }

                L1_error += error;
              }
              REQUIRE(parallel::allReduceMax(L1_error) == Catch::Approx(0).margin(1E-13));
            }
          }
        }
      }

      SECTION("R^3x3 data")
      {
        using R3x3 = TinyMatrix<3, 3>;

        for (auto named_mesh : MeshDataBaseForTests::get().all1DMeshes()) {
          SECTION(named_mesh.name())
          {
            auto p_mesh = named_mesh.mesh()->get<Mesh<1>>();
            auto& mesh  = *p_mesh;

            auto A_exact = [](const R1& X) -> R3x3 {
              const double x  = X[0];
              const double x2 = x * x;
              return R3x3{+2.3 + 1.7 * x + 2.9 * x2,   //
                          -1.7 + 2.1 * x + 1.7 * x2,   //
                          +1.4 - 0.6 * x - 0.5 * x2,   //
                                                       //
                          +2.4 - 2.3 * x + 2.3 * x2,   //
                          -0.2 + 3.1 * x - 1.9 * x2,   //
                          -3.2 - 3.6 * x - 0.3 * x2,   //
                                                       //
                          -4.1 + 3.1 * x - 1.3 * x2,   //
                          +0.8 + 2.9 * x + 2.1 * x2,   //
                          -1.6 + 2.3 * x + 0.8 * x2};
            };

            auto xr = mesh.xr();
            auto Vj = MeshDataManager::instance().getMeshData(mesh).Vj();

            auto cell_to_node_matrix = mesh.connectivity().cellToNodeMatrix();
            DiscreteFunctionP0<R3x3> Ah{p_mesh};

            parallel_for(
              mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
                auto cell_node_list = cell_to_node_matrix[cell_id];
                R3x3 value          = zero;
                LineTransformation<1> T{xr[cell_node_list[0]], xr[cell_node_list[1]]};

                for (size_t i_q = 0; i_q < qf.numberOfPoints(); ++i_q) {
                  value += qf.weight(i_q) * T.jacobianDeterminant() * A_exact(T(qf.point(i_q)));
                }

                Ah[cell_id] = 1. / Vj[cell_id] * value;
              });

            auto reconstructions = PolynomialReconstruction{descriptor}.build(Ah);

            auto dpk_Ah = reconstructions[0]->get<DiscreteFunctionDPk<1, const R3x3>>();

            {
              double L1_error = 0;
              for (CellId cell_id = 0; cell_id < mesh.numberOfCells(); ++cell_id) {
                const auto cell_node_list = cell_to_node_matrix[cell_id];
                const auto dpkj           = dpk_Ah[cell_id];

                LineTransformation<1> T{xr[cell_node_list[0]], xr[cell_node_list[1]]};

                double error = 0;

                for (size_t i_q = 0; i_q < qf.numberOfPoints(); ++i_q) {
                  R3x3 diff =
                    qf.weight(i_q) * T.jacobianDeterminant() * (A_exact(T(qf.point(i_q))) - dpkj(T(qf.point(i_q))));
                  for (size_t i = 0; i < diff.numberOfRows(); ++i) {
                    for (size_t j = 0; j < diff.numberOfColumns(); ++j) {
                      error += std::abs(diff(i, j));
                    }
                  }
                }

                L1_error += error;
              }
              REQUIRE(parallel::allReduceMax(L1_error) == Catch::Approx(0).margin(1E-13));
            }
          }
        }
      }

      SECTION("R vector data")
      {
        using R3 = TinyVector<3>;

        for (auto named_mesh : MeshDataBaseForTests::get().all1DMeshes()) {
          SECTION(named_mesh.name())
          {
            auto p_mesh = named_mesh.mesh()->get<Mesh<1>>();
            auto& mesh  = *p_mesh;

            auto u_exact = [](const R1& X) -> R3 {
              const double x = X[0];

              return R3{2.3 + 1.7 * x - 3.2 * x * x,   //
                        7 + 5 * x + 3 * x * x,         //
                        -4 + 3.5 * x + 2.7 * x * x};
            };

            auto xr = mesh.xr();
            auto Vj = MeshDataManager::instance().getMeshData(mesh).Vj();

            auto cell_to_node_matrix = mesh.connectivity().cellToNodeMatrix();

            DiscreteFunctionP0Vector<double> uh{p_mesh, 3};

            parallel_for(
              mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
                auto cell_node_list = cell_to_node_matrix[cell_id];
                R3 value            = zero;
                LineTransformation<1> T{xr[cell_node_list[0]], xr[cell_node_list[1]]};

                for (size_t i_q = 0; i_q < qf.numberOfPoints(); ++i_q) {
                  value += qf.weight(i_q) * T.jacobianDeterminant() * u_exact(T(qf.point(i_q)));
                }

                value *= 1. / Vj[cell_id];

                for (size_t i = 0; i < value.dimension(); ++i) {
                  uh[cell_id][i] = value[i];
                }
              });

            auto reconstructions = PolynomialReconstruction{descriptor}.build(uh);

            auto dpk_uh = reconstructions[0]->get<DiscreteFunctionDPkVector<1, const double>>();

            {
              double L1_error = 0;
              for (CellId cell_id = 0; cell_id < mesh.numberOfCells(); ++cell_id) {
                const auto cell_node_list = cell_to_node_matrix[cell_id];

                LineTransformation<1> T{xr[cell_node_list[0]], xr[cell_node_list[1]]};

                double error = 0;

                for (size_t i_q = 0; i_q < qf.numberOfPoints(); ++i_q) {
                  for (size_t l = 0; l < dpk_uh.numberOfComponents(); ++l) {
                    error += std::abs(qf.weight(i_q) * T.jacobianDeterminant() *
                                      (u_exact(T(qf.point(i_q)))[l] - dpk_uh(cell_id, l)(T(qf.point(i_q)))));
                  }
                }

                L1_error += error;
              }
              REQUIRE(parallel::allReduceMax(L1_error) == Catch::Approx(0).margin(1E-13));
            }
          }
        }
      }

      SECTION("list of various types")
      {
        using R3x3 = TinyMatrix<3>;
        using R3   = TinyVector<3>;

        for (auto named_mesh : MeshDataBaseForTests::get().all1DMeshes()) {
          SECTION(named_mesh.name())
          {
            auto p_mesh = named_mesh.mesh()->get<Mesh<1>>();
            auto& mesh  = *p_mesh;

            auto f_exact = [](const R1& x) { return 2.3 + 1.7 * x[0] - 3.2 * x[0] * x[0]; };

            auto u_exact = [](const R1& X) -> R3 {
              const double x = X[0];

              return R3{2.3 + 1.7 * x - 3.2 * x * x,   //
                        7 + 5 * x + 3 * x * x,         //
                        -4 + 3.5 * x + 2.7 * x * x};
            };

            auto A_exact = [](const R1& X) -> R3x3 {
              const double x  = X[0];
              const double x2 = x * x;
              return R3x3{+2.3 + 1.7 * x + 2.9 * x2,   //
                          -1.7 + 2.1 * x + 1.7 * x2,   //
                          +1.4 - 0.6 * x - 0.5 * x2,   //
                                                       //
                          +2.4 - 2.3 * x + 2.3 * x2,   //
                          -0.2 + 3.1 * x - 1.9 * x2,   //
                          -3.2 - 3.6 * x - 0.3 * x2,   //
                                                       //
                          -4.1 + 3.1 * x - 1.3 * x2,   //
                          +0.8 + 2.9 * x + 2.1 * x2,   //
                          -1.6 + 2.3 * x + 0.8 * x2};
            };

            auto xr = mesh.xr();
            auto Vj = MeshDataManager::instance().getMeshData(mesh).Vj();

            auto cell_to_node_matrix = mesh.connectivity().cellToNodeMatrix();

            DiscreteFunctionP0Vector<double> vh{p_mesh, 3};

            parallel_for(
              mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
                auto cell_node_list = cell_to_node_matrix[cell_id];
                R3 value            = zero;
                LineTransformation<1> T{xr[cell_node_list[0]], xr[cell_node_list[1]]};

                for (size_t i_q = 0; i_q < qf.numberOfPoints(); ++i_q) {
                  value += qf.weight(i_q) * T.jacobianDeterminant() * u_exact(T(qf.point(i_q)));
                }

                value *= 1. / Vj[cell_id];

                for (size_t i = 0; i < value.dimension(); ++i) {
                  vh[cell_id][i] = value[i];
                }
              });

            DiscreteFunctionP0<double> fh{p_mesh};

            parallel_for(
              mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
                auto cell_node_list = cell_to_node_matrix[cell_id];
                double value        = 0;
                LineTransformation<1> T{xr[cell_node_list[0]], xr[cell_node_list[1]]};

                for (size_t i_q = 0; i_q < qf.numberOfPoints(); ++i_q) {
                  value += qf.weight(i_q) * f_exact(T(qf.point(i_q))) * T.jacobianDeterminant();
                }

                fh[cell_id] = value / Vj[cell_id];
              });

            DiscreteFunctionP0<R3x3> Ah{p_mesh};

            parallel_for(
              mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
                auto cell_node_list = cell_to_node_matrix[cell_id];
                R3x3 value          = zero;
                LineTransformation<1> T{xr[cell_node_list[0]], xr[cell_node_list[1]]};

                for (size_t i_q = 0; i_q < qf.numberOfPoints(); ++i_q) {
                  value += qf.weight(i_q) * T.jacobianDeterminant() * A_exact(T(qf.point(i_q)));
                }

                Ah[cell_id] = 1. / Vj[cell_id] * value;
              });

            auto reconstructions = PolynomialReconstruction{descriptor}.build(Ah, vh, fh);

            auto dpk_Ah = reconstructions[0]->get<DiscreteFunctionDPk<1, const R3x3>>();
            auto dpk_vh = reconstructions[1]->get<DiscreteFunctionDPkVector<1, const double>>();
            auto dpk_fh = reconstructions[2]->get<DiscreteFunctionDPk<1, const double>>();

            {
              double L1_error = 0;
              for (CellId cell_id = 0; cell_id < mesh.numberOfCells(); ++cell_id) {
                const auto cell_node_list = cell_to_node_matrix[cell_id];

                LineTransformation<1> T{xr[cell_node_list[0]], xr[cell_node_list[1]]};

                double error = 0;

                for (size_t i_q = 0; i_q < qf.numberOfPoints(); ++i_q) {
                  for (size_t l = 0; l < dpk_vh.numberOfComponents(); ++l) {
                    error += std::abs(qf.weight(i_q) * T.jacobianDeterminant() *
                                      (u_exact(T(qf.point(i_q)))[l] - dpk_vh(cell_id, l)(T(qf.point(i_q)))));
                  }
                }

                L1_error += error;
              }
              REQUIRE(parallel::allReduceMax(L1_error) == Catch::Approx(0).margin(1E-13));
            }

            {
              double L1_error = 0;
              for (CellId cell_id = 0; cell_id < mesh.numberOfCells(); ++cell_id) {
                const auto cell_node_list = cell_to_node_matrix[cell_id];
                const auto dpkj           = dpk_fh[cell_id];

                LineTransformation<1> T{xr[cell_node_list[0]], xr[cell_node_list[1]]};

                double error = 0;

                for (size_t i_q = 0; i_q < qf.numberOfPoints(); ++i_q) {
                  error += std::abs(qf.weight(i_q) * (f_exact(T(qf.point(i_q))) - dpkj(T(qf.point(i_q)))) *
                                    T.jacobianDeterminant());
                }

                L1_error += error;
              }
              REQUIRE(parallel::allReduceMax(L1_error) == Catch::Approx(0).margin(1E-13));
            }

            {
              double L1_error = 0;
              for (CellId cell_id = 0; cell_id < mesh.numberOfCells(); ++cell_id) {
                const auto cell_node_list = cell_to_node_matrix[cell_id];
                const auto dpkj           = dpk_Ah[cell_id];

                LineTransformation<1> T{xr[cell_node_list[0]], xr[cell_node_list[1]]};

                double error = 0;

                for (size_t i_q = 0; i_q < qf.numberOfPoints(); ++i_q) {
                  R3x3 diff =
                    qf.weight(i_q) * T.jacobianDeterminant() * (A_exact(T(qf.point(i_q))) - dpkj(T(qf.point(i_q))));
                  for (size_t i = 0; i < diff.numberOfRows(); ++i) {
                    for (size_t j = 0; j < diff.numberOfColumns(); ++j) {
                      error += std::abs(diff(i, j));
                    }
                  }
                }

                L1_error += error;
              }
              REQUIRE(parallel::allReduceMax(L1_error) == Catch::Approx(0).margin(1E-13));
            }
          }
        }
      }
    }

    SECTION("2D")
    {
      using R2 = TinyVector<2>;

      auto integrate_in_cell = [](const CellType& cell_type, const auto& cell_node_list, const auto& xr,
                                  const auto& exact, auto& value) {
        switch (cell_type) {
        case CellType::Triangle: {
          TriangleTransformation<2> T{xr[cell_node_list[0]], xr[cell_node_list[1]], xr[cell_node_list[2]]};
          QuadratureFormula<2> qf = QuadratureManager::instance().getTriangleFormula(GaussQuadratureDescriptor{2});
          for (size_t i_q = 0; i_q < qf.numberOfPoints(); ++i_q) {
            value += qf.weight(i_q) * T.jacobianDeterminant() * exact(T(qf.point(i_q)));
          }
          break;
        }
        case CellType::Quadrangle: {
          SquareTransformation<2> T{xr[cell_node_list[0]], xr[cell_node_list[1]], xr[cell_node_list[2]],
                                    xr[cell_node_list[3]]};
          QuadratureFormula<2> qf =
            QuadratureManager::instance().getSquareFormula(GaussLegendreQuadratureDescriptor{3});
          for (size_t i_q = 0; i_q < qf.numberOfPoints(); ++i_q) {
            const R2 x_hat = qf.point(i_q);
            const R2 x     = T(x_hat);
            value += qf.weight(i_q) * T.jacobianDeterminant(x_hat) * exact(x);
          }
          break;
        }
        default: {
          throw UnexpectedError("invalid cell type");
        }
        }
      };

      auto compute_L1_error = [](const auto& cell_type, const auto& mesh, const auto& exact, const auto& dpk) {
        auto cell_to_node_matrix = mesh.connectivity().cellToNodeMatrix();
        const auto& xr           = mesh.xr();

        using DataType = typename std::decay_t<decltype(dpk)>::data_type;

        auto sum_abs = [](const DataType& diff) -> double {
          if constexpr (std::is_arithmetic_v<DataType>) {
            return std::abs(diff);
          } else if constexpr (is_tiny_vector_v<DataType>) {
            double sum = 0;
            for (size_t i = 0; i < diff.dimension(); ++i) {
              sum += std::abs(diff[i]);
            }
            return sum;
          } else if constexpr (is_tiny_matrix_v<DataType>) {
            double sum = 0;
            for (size_t i = 0; i < diff.numberOfRows(); ++i) {
              for (size_t j = 0; j < diff.numberOfRows(); ++j) {
                sum += std::abs(diff(i, j));
              }
            }
            return sum;
          } else {
            throw UnexpectedError("unexpected value type");
          }
        };

        double L1_error = 0;
        for (CellId cell_id = 0; cell_id < mesh.numberOfCells(); ++cell_id) {
          const auto cell_node_list = cell_to_node_matrix[cell_id];
          const auto dpkj           = dpk[cell_id];

          double error = 0;

          switch (cell_type[cell_id]) {
          case CellType::Triangle: {
            TriangleTransformation<2> T{xr[cell_node_list[0]], xr[cell_node_list[1]], xr[cell_node_list[2]]};
            QuadratureFormula<2> qf = QuadratureManager::instance().getTriangleFormula(GaussQuadratureDescriptor{2});
            for (size_t i_q = 0; i_q < qf.numberOfPoints(); ++i_q) {
              const R2 x    = T(qf.point(i_q));
              DataType diff = qf.weight(i_q) * T.jacobianDeterminant() * (exact(x) - dpkj(x));
              error += sum_abs(diff);
            }
            break;
          }
          case CellType::Quadrangle: {
            SquareTransformation<2> T{xr[cell_node_list[0]], xr[cell_node_list[1]], xr[cell_node_list[2]],
                                      xr[cell_node_list[3]]};
            QuadratureFormula<2> qf =
              QuadratureManager::instance().getSquareFormula(GaussLegendreQuadratureDescriptor{3});
            for (size_t i_q = 0; i_q < qf.numberOfPoints(); ++i_q) {
              const R2 x_hat = qf.point(i_q);
              const R2 x     = T(x_hat);
              DataType diff  = qf.weight(i_q) * T.jacobianDeterminant(x_hat) * (exact(x) - dpkj(x));
              error += sum_abs(diff);
            }
            break;
          }
          default: {
            throw UnexpectedError("invalid cell type");
          }
          }

          L1_error += error;
        }
        return L1_error;
      };

      std::vector<PolynomialReconstructionDescriptor> descriptor_list =
        {PolynomialReconstructionDescriptor{IntegrationMethodType::element, 2},
         PolynomialReconstructionDescriptor{IntegrationMethodType::boundary, 2}};

      for (auto descriptor : descriptor_list) {
        SECTION(name(descriptor.integrationMethodType()))
        {
          SECTION("R data")
          {
            for (auto named_mesh : MeshDataBaseForTests::get().all2DMeshes()) {
              SECTION(named_mesh.name())
              {
                auto p_mesh = named_mesh.mesh()->get<Mesh<2>>();
                auto& mesh  = *p_mesh;

                auto f_exact = [](const R2& x) {
                  return 2.3 + 1.7 * x[0] + 0.2 * x[1] - 3.2 * x[0] * x[0] + 1.3 * x[0] * x[1] - 1.4 * x[1] * x[1];
                };

                auto xr = mesh.xr();
                auto Vj = MeshDataManager::instance().getMeshData(mesh).Vj();

                auto cell_type           = mesh.connectivity().cellType();
                auto cell_to_node_matrix = mesh.connectivity().cellToNodeMatrix();

                DiscreteFunctionP0<double> fh{p_mesh};

                parallel_for(
                  mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
                    auto cell_node_list = cell_to_node_matrix[cell_id];
                    double value        = 0;
                    integrate_in_cell(cell_type[cell_id], cell_node_list, xr, f_exact, value);
                    fh[cell_id] = value / Vj[cell_id];
                  });

                auto reconstructions = PolynomialReconstruction{descriptor}.build(fh);

                auto dpk_fh = reconstructions[0]->get<DiscreteFunctionDPk<2, const double>>();

                double L1_error = compute_L1_error(cell_type, mesh, f_exact, dpk_fh);
                REQUIRE(parallel::allReduceMax(L1_error) == Catch::Approx(0).margin(1E-13));
              }
            }
          }

          SECTION("R^3 data")
          {
            using R3 = TinyVector<3>;

            for (auto named_mesh : MeshDataBaseForTests::get().all2DMeshes()) {
              SECTION(named_mesh.name())
              {
                auto p_mesh = named_mesh.mesh()->get<Mesh<2>>();
                auto& mesh  = *p_mesh;

                auto u_exact = [](const R2& X) -> R3 {
                  const double x  = X[0];
                  const double y  = X[1];
                  const double x2 = x * x;
                  const double xy = x * y;
                  const double y2 = y * y;

                  return R3{+2.3 + 1.7 * x + 1.3 * y - 3.2 * x2 + 1.6 * xy + 0.9 * y2,   //
                            +7.0 + 5.0 * x - 2.4 * y + 3.0 * x2 - 0.8 * xy + 1.1 * y2,   //
                            -4.0 + 3.5 * x - 0.7 * y + 2.7 * x2 - 2.1 * xy - 1.8 * y2};
                };

                auto xr = mesh.xr();
                auto Vj = MeshDataManager::instance().getMeshData(mesh).Vj();

                auto cell_type           = mesh.connectivity().cellType();
                auto cell_to_node_matrix = mesh.connectivity().cellToNodeMatrix();

                DiscreteFunctionP0<R3> uh{p_mesh};

                parallel_for(
                  mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
                    auto cell_node_list = cell_to_node_matrix[cell_id];
                    R3 value            = zero;
                    integrate_in_cell(cell_type[cell_id], cell_node_list, xr, u_exact, value);
                    uh[cell_id] = 1. / Vj[cell_id] * value;
                  });

                auto reconstructions = PolynomialReconstruction{descriptor}.build(uh);

                auto dpk_uh = reconstructions[0]->get<DiscreteFunctionDPk<2, const R3>>();

                double L1_error = compute_L1_error(cell_type, mesh, u_exact, dpk_uh);
                REQUIRE(parallel::allReduceMax(L1_error) == Catch::Approx(0).margin(1E-13));
              }
            }
          }

          SECTION("R^2x2 data")
          {
            using R2x2 = TinyMatrix<2, 2>;

            for (auto named_mesh : MeshDataBaseForTests::get().all2DMeshes()) {
              SECTION(named_mesh.name())
              {
                auto p_mesh = named_mesh.mesh()->get<Mesh<2>>();
                auto& mesh  = *p_mesh;

                auto A_exact = [](const R2& X) -> R2x2 {
                  const double x  = X[0];
                  const double y  = X[1];
                  const double x2 = x * x;
                  const double xy = x * y;
                  const double y2 = y * y;

                  return R2x2{+2.3 + 1.7 * x + 1.2 * y + 1.1 * x2 + 0.7 * xy - 0.8 * y2,   //
                              -1.7 + 2.1 * x - 2.2 * y - 1.9 * x2 + 0.2 * xy + 1.2 * y2,   //
                              +1.4 - 0.6 * x - 2.1 * y + 0.3 * x2 - 3.1 * xy + 1.3 * y2,   //
                              +2.4 - 2.3 * x + 1.3 * y - 0.8 * x2 + 1.2 * xy - 2.0 * y2};
                };

                auto xr = mesh.xr();
                auto Vj = MeshDataManager::instance().getMeshData(mesh).Vj();

                auto cell_type           = mesh.connectivity().cellType();
                auto cell_to_node_matrix = mesh.connectivity().cellToNodeMatrix();

                DiscreteFunctionP0<R2x2> Ah{p_mesh};

                parallel_for(
                  mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
                    auto cell_node_list = cell_to_node_matrix[cell_id];
                    R2x2 value          = zero;
                    integrate_in_cell(cell_type[cell_id], cell_node_list, xr, A_exact, value);
                    Ah[cell_id] = 1. / Vj[cell_id] * value;
                  });

                auto reconstructions = PolynomialReconstruction{descriptor}.build(Ah);

                auto dpk_Ah = reconstructions[0]->get<DiscreteFunctionDPk<2, const R2x2>>();

                double L1_error = compute_L1_error(cell_type, mesh, A_exact, dpk_Ah);
                REQUIRE(parallel::allReduceMax(L1_error) == Catch::Approx(0).margin(1E-13));
              }
            }
          }
        }
      }
    }

    SECTION("3D")
    {
      PolynomialReconstructionDescriptor descriptor{IntegrationMethodType::element, 2};

      using R3 = TinyVector<3>;

      auto integrate_in_cell = [](const CellType& cell_type, const auto& cell_node_list, const auto& xr,
                                  const auto& exact, auto& value) {
        switch (cell_type) {
        case CellType::Tetrahedron: {
          TetrahedronTransformation T{xr[cell_node_list[0]], xr[cell_node_list[1]], xr[cell_node_list[2]],
                                      xr[cell_node_list[3]]};
          QuadratureFormula<3> qf = QuadratureManager::instance().getTetrahedronFormula(GaussQuadratureDescriptor{2});
          for (size_t i_q = 0; i_q < qf.numberOfPoints(); ++i_q) {
            const R3 x_hat = qf.point(i_q);
            const R3 x     = T(x_hat);
            value += qf.weight(i_q) * T.jacobianDeterminant() * exact(x);
          }
          break;
        }
        case CellType::Pyramid: {
          PyramidTransformation T{xr[cell_node_list[0]], xr[cell_node_list[1]], xr[cell_node_list[2]],
                                  xr[cell_node_list[3]], xr[cell_node_list[4]]};
          QuadratureFormula<3> qf = QuadratureManager::instance().getPyramidFormula(GaussQuadratureDescriptor{3});
          for (size_t i_q = 0; i_q < qf.numberOfPoints(); ++i_q) {
            const R3 x_hat = qf.point(i_q);
            const R3 x     = T(x_hat);
            value += qf.weight(i_q) * T.jacobianDeterminant(x_hat) * exact(x);
          }
          break;
        }
        case CellType::Prism: {
          PrismTransformation T{xr[cell_node_list[0]], xr[cell_node_list[1]], xr[cell_node_list[2]],
                                xr[cell_node_list[3]], xr[cell_node_list[4]], xr[cell_node_list[5]]};
          QuadratureFormula<3> qf = QuadratureManager::instance().getPrismFormula(GaussQuadratureDescriptor{3});
          for (size_t i_q = 0; i_q < qf.numberOfPoints(); ++i_q) {
            const R3 x_hat = qf.point(i_q);
            const R3 x     = T(x_hat);
            value += qf.weight(i_q) * T.jacobianDeterminant(x_hat) * exact(x);
          }
          break;
        }
        case CellType::Hexahedron: {
          CubeTransformation T{xr[cell_node_list[0]], xr[cell_node_list[1]], xr[cell_node_list[2]],
                               xr[cell_node_list[3]], xr[cell_node_list[4]], xr[cell_node_list[5]],
                               xr[cell_node_list[6]], xr[cell_node_list[7]]};
          QuadratureFormula<3> qf = QuadratureManager::instance().getCubeFormula(GaussLegendreQuadratureDescriptor{3});
          for (size_t i_q = 0; i_q < qf.numberOfPoints(); ++i_q) {
            const R3 x_hat = qf.point(i_q);
            const R3 x     = T(x_hat);
            value += qf.weight(i_q) * T.jacobianDeterminant(x_hat) * exact(x);
          }
          break;
        }
        default: {
          throw UnexpectedError("invalid cell type");
        }
        }
      };

      auto compute_L1_error = [](const auto& cell_type, const auto& mesh, const auto& exact, const auto& dpk) {
        auto cell_to_node_matrix = mesh.connectivity().cellToNodeMatrix();
        const auto& xr           = mesh.xr();

        using DataType = typename std::decay_t<decltype(dpk)>::data_type;

        auto sum_abs = [](const DataType& diff) -> double {
          if constexpr (std::is_arithmetic_v<DataType>) {
            return std::abs(diff);
          } else if constexpr (is_tiny_vector_v<DataType>) {
            double sum = 0;
            for (size_t i = 0; i < diff.dimension(); ++i) {
              sum += std::abs(diff[i]);
            }
            return sum;
          } else if constexpr (is_tiny_matrix_v<DataType>) {
            double sum = 0;
            for (size_t i = 0; i < diff.numberOfRows(); ++i) {
              for (size_t j = 0; j < diff.numberOfRows(); ++j) {
                sum += std::abs(diff(i, j));
              }
            }
            return sum;
          } else {
            throw UnexpectedError("unexpected value type");
          }
        };

        double L1_error = 0;
        for (CellId cell_id = 0; cell_id < mesh.numberOfCells(); ++cell_id) {
          const auto cell_node_list = cell_to_node_matrix[cell_id];
          const auto dpkj           = dpk[cell_id];

          double error = 0;

          switch (cell_type[cell_id]) {
          case CellType::Tetrahedron: {
            TetrahedronTransformation T{xr[cell_node_list[0]], xr[cell_node_list[1]], xr[cell_node_list[2]],
                                        xr[cell_node_list[3]]};
            QuadratureFormula<3> qf = QuadratureManager::instance().getTetrahedronFormula(GaussQuadratureDescriptor{2});
            for (size_t i_q = 0; i_q < qf.numberOfPoints(); ++i_q) {
              const R3 x_hat = qf.point(i_q);
              const R3 x     = T(x_hat);
              DataType diff  = qf.weight(i_q) * T.jacobianDeterminant() * (exact(x) - dpkj(x));
              error += sum_abs(diff);
            }
            break;
          }
          case CellType::Pyramid: {
            PyramidTransformation T{xr[cell_node_list[0]], xr[cell_node_list[1]], xr[cell_node_list[2]],
                                    xr[cell_node_list[3]], xr[cell_node_list[4]]};
            QuadratureFormula<3> qf = QuadratureManager::instance().getPyramidFormula(GaussQuadratureDescriptor{3});
            for (size_t i_q = 0; i_q < qf.numberOfPoints(); ++i_q) {
              const R3 x_hat = qf.point(i_q);
              const R3 x     = T(x_hat);
              DataType diff  = qf.weight(i_q) * T.jacobianDeterminant(x_hat) * (exact(x) - dpkj(x));
              error += sum_abs(diff);
            }
            break;
          }
          case CellType::Prism: {
            PrismTransformation T{xr[cell_node_list[0]], xr[cell_node_list[1]], xr[cell_node_list[2]],
                                  xr[cell_node_list[3]], xr[cell_node_list[4]], xr[cell_node_list[5]]};
            QuadratureFormula<3> qf = QuadratureManager::instance().getPrismFormula(GaussQuadratureDescriptor{3});
            for (size_t i_q = 0; i_q < qf.numberOfPoints(); ++i_q) {
              const R3 x_hat = qf.point(i_q);
              const R3 x     = T(x_hat);
              DataType diff  = qf.weight(i_q) * T.jacobianDeterminant(x_hat) * (exact(x) - dpkj(x));
              error += sum_abs(diff);
            }
            break;
          }
          case CellType::Hexahedron: {
            CubeTransformation T{xr[cell_node_list[0]], xr[cell_node_list[1]], xr[cell_node_list[2]],
                                 xr[cell_node_list[3]], xr[cell_node_list[4]], xr[cell_node_list[5]],
                                 xr[cell_node_list[6]], xr[cell_node_list[7]]};
            QuadratureFormula<3> qf =
              QuadratureManager::instance().getCubeFormula(GaussLegendreQuadratureDescriptor{3});
            for (size_t i_q = 0; i_q < qf.numberOfPoints(); ++i_q) {
              const R3 x_hat = qf.point(i_q);
              const R3 x     = T(x_hat);
              DataType diff  = qf.weight(i_q) * T.jacobianDeterminant(x_hat) * (exact(x) - dpkj(x));
              error += sum_abs(diff);
            }
            break;
          }
          default: {
            throw UnexpectedError("invalid cell type");
          }
          }

          L1_error += error;
        }
        return L1_error;
      };

      SECTION("R data")
      {
        for (auto named_mesh : MeshDataBaseForTests::get().all3DMeshes()) {
          SECTION(named_mesh.name())
          {
            auto p_mesh  = named_mesh.mesh()->get<Mesh<3>>();
            auto& mesh   = *p_mesh;
            auto f_exact = [](const R3& X) {
              const double x  = X[0];
              const double y  = X[1];
              const double z  = X[2];
              const double x2 = x * x;
              const double xy = x * y;
              const double xz = x * z;
              const double y2 = y * y;
              const double yz = y * z;
              const double z2 = z * z;

              return 2.3 + 1.7 * x + 0.2 * y + 1.4 * z - 3.2 * x2 + 1.3 * xy - 1.6 * xz - 1.4 * y2 + 2 * yz - 1.8 * z2;
            };

            auto xr = mesh.xr();
            auto Vj = MeshDataManager::instance().getMeshData(mesh).Vj();

            auto cell_type           = mesh.connectivity().cellType();
            auto cell_to_node_matrix = mesh.connectivity().cellToNodeMatrix();

            DiscreteFunctionP0<double> fh{p_mesh};

            parallel_for(
              mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
                auto cell_node_list = cell_to_node_matrix[cell_id];
                double value        = 0;
                integrate_in_cell(cell_type[cell_id], cell_node_list, xr, f_exact, value);
                fh[cell_id] = value / Vj[cell_id];
              });

            auto reconstructions = PolynomialReconstruction{descriptor}.build(fh);

            auto dpk_fh = reconstructions[0]->get<DiscreteFunctionDPk<3, const double>>();

            double L1_error = compute_L1_error(cell_type, mesh, f_exact, dpk_fh);
            REQUIRE(parallel::allReduceMax(L1_error) == Catch::Approx(0).margin(1E-13));
          }
        }
      }

      SECTION("R^3 data")
      {
        for (auto named_mesh : MeshDataBaseForTests::get().all3DMeshes()) {
          SECTION(named_mesh.name())
          {
            auto p_mesh = named_mesh.mesh()->get<Mesh<3>>();
            auto& mesh  = *p_mesh;

            auto u_exact = [](const R3& X) -> R3 {
              const double x  = X[0];
              const double y  = X[1];
              const double z  = X[2];
              const double x2 = x * x;
              const double xy = x * y;
              const double xz = x * z;
              const double y2 = y * y;
              const double yz = y * z;
              const double z2 = z * z;

              return R3{+2.3 + 1.7 * x - 2.2 * y + 1.8 * z                                     //
                          + 0.3 * x2 - 1.3 * xy + 2.7 * xz + 0.7 * y2 + 2.0 * yz - 1.2 * z2,   //
                        +1.4 - 0.6 * x + 1.3 * y - 3.7 * z                                     //
                          + 3.1 * x2 - 2.4 * xy - 1.5 * xz - 1.9 * y2 + 0.2 * yz + 0.9 * z2,   //
                        -0.2 + 3.1 * x - 1.1 * y + 1.9 * z                                     //
                          - 1.3 * x2 + 2.8 * xy - 2.1 * xz + 3.2 * y2 - 2.1 * yz + 1.6 * z2};
            };

            auto xr = mesh.xr();
            auto Vj = MeshDataManager::instance().getMeshData(mesh).Vj();

            auto cell_type           = mesh.connectivity().cellType();
            auto cell_to_node_matrix = mesh.connectivity().cellToNodeMatrix();

            DiscreteFunctionP0<R3> uh{p_mesh};

            parallel_for(
              mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
                auto cell_node_list = cell_to_node_matrix[cell_id];
                R3 value            = zero;
                integrate_in_cell(cell_type[cell_id], cell_node_list, xr, u_exact, value);
                uh[cell_id] = 1. / Vj[cell_id] * value;
              });

            auto reconstructions = PolynomialReconstruction{descriptor}.build(uh);

            auto dpk_uh     = reconstructions[0]->get<DiscreteFunctionDPk<3, const R3>>();
            double L1_error = compute_L1_error(cell_type, mesh, u_exact, dpk_uh);
            REQUIRE(parallel::allReduceMax(L1_error) == Catch::Approx(0).margin(1E-13));
          }
        }
      }

      SECTION("R^2x2 data")
      {
        using R2x2 = TinyMatrix<2, 2>;

        for (auto named_mesh : MeshDataBaseForTests::get().all3DMeshes()) {
          SECTION(named_mesh.name())
          {
            auto p_mesh = named_mesh.mesh()->get<Mesh<3>>();
            auto& mesh  = *p_mesh;

            auto A_exact = [](const R3& X) -> R2x2 {
              const double x  = X[0];
              const double y  = X[1];
              const double z  = X[2];
              const double x2 = x * x;
              const double xy = x * y;
              const double xz = x * z;
              const double y2 = y * y;
              const double yz = y * z;
              const double z2 = z * z;

              return R2x2{+2.3 + 1.7 * x - 2.2 * y + 1.8 * z                                     //
                            + 0.3 * x2 - 1.3 * xy + 2.7 * xz + 0.7 * y2 + 2.0 * yz - 1.2 * z2,   //
                          +1.4 - 0.6 * x + 1.3 * y - 3.7 * z                                     //
                            + 3.1 * x2 - 2.4 * xy - 1.5 * xz - 1.9 * y2 + 0.2 * yz + 0.9 * z2,   //
                                                                                                 //
                          -0.2 + 3.1 * x - 1.1 * y + 1.9 * z                                     //
                            - 1.3 * x2 + 2.8 * xy - 2.1 * xz + 3.2 * y2 - 2.1 * yz + 1.6 * z2,   //
                          +0.9 - 1.3 * x + 2.1 * y + 3.1 * z                                     //
                            + 1.1 * x2 + 1.8 * xy + 1.2 * xz - 2.3 * y2 + 3.3 * yz - 1.2 * z2};
            };

            auto xr = mesh.xr();
            auto Vj = MeshDataManager::instance().getMeshData(mesh).Vj();

            auto cell_type           = mesh.connectivity().cellType();
            auto cell_to_node_matrix = mesh.connectivity().cellToNodeMatrix();

            DiscreteFunctionP0<R2x2> Ah{p_mesh};

            parallel_for(
              mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
                auto cell_node_list = cell_to_node_matrix[cell_id];
                R2x2 value          = zero;
                integrate_in_cell(cell_type[cell_id], cell_node_list, xr, A_exact, value);
                Ah[cell_id] = 1. / Vj[cell_id] * value;
              });

            descriptor.setRowWeighting(false);
            auto reconstructions = PolynomialReconstruction{descriptor}.build(Ah);

            auto dpk_Ah = reconstructions[0]->get<DiscreteFunctionDPk<3, const R2x2>>();
          }
        }
      }
    }
  }
}
